@@ -12,7 +12,7 @@ import textwrap
 version = csrbox.__version__
 logger = logging.getLogger(__name__)
 
-def csr_gen (isa_file, grp_file, work_dir, logging=False):
+def csr_gen (isa_file, grp_file, custom_file, custom_attr_file, work_dir, logging=False):
     '''
         Top level function which validates the sanity of the grouping yaml 
         followed by calls to subfunctions which generated the corresponding 
@@ -34,9 +34,14 @@ def csr_gen (isa_file, grp_file, work_dir, logging=False):
     
     isa_yaml = utils.load_yaml(isa_file)
     grp_yaml = utils.load_yaml(grp_file)
+
+    if custom_file :
+        custom_yaml= utils.load_yaml(custom_file)
+        isa_yaml['hart0'].update(custom_yaml['hart0'])
+
     
     csrgrp_check.csrgrp_check(isa_yaml, grp_yaml, logging)
-    attr_yaml = csrgrp_check.calculate_size(isa_yaml)
+    attr_yaml = csrgrp_check.calculate_size(isa_yaml, custom_attr_file)
     csr_define(isa_yaml , work_dir , 'csrbox.defines', logging)
     csr_decoder(isa_yaml, work_dir , 'csrbox_decoder.bsv', logging)
     csr_types(isa_yaml  , work_dir , 'csr_types.bsv', logging)
@@ -99,6 +104,9 @@ def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, logging=False):
         :type attr_yaml: dict 
         :type work_dir: str
     '''
+
+    global physical_addr_sz
+
     rvxlen = 'rv' + str(max(isa_yaml['hart0']['supported_xlen']))
     xlen = str(max(isa_yaml['hart0']['supported_xlen']))
 
@@ -109,12 +117,13 @@ def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, logging=False):
     max_len = len(grp_yaml)
 
     for grp, elems in grp_yaml.items():
-        if 'grp'+str(max_len) == grp:
-            last = True
-        else:
-            last = False
-        for node in isa_yaml['hart0']:
-            if isinstance(isa_yaml['hart0'][node], dict):
+        if elems != None:
+          if 'grp'+str(max_len) == grp:
+              last = True
+          else:
+              last = False
+          for node in isa_yaml['hart0']:
+             if isinstance(isa_yaml['hart0'][node], dict):
                 if 'description' in isa_yaml['hart0'][node]:
                     csr_name = node
                     name_str = str(csr_name)
@@ -148,7 +157,8 @@ def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, logging=False):
         bsvstr += (end_module_temp)
         bsvstr +=(end_package_temp)
         bsvfile.write(bsvstr)
-        station_isa.clear()   
+        station_isa.clear()  
+    physical_addr_sz = isa_yaml['hart0']['physical_addr_sz']
     bsvstr = csr_top(csr_isa, grp_yaml, attr_yaml, rvxlen,
              'csrbox', logging)
     filename = 'csrbox.bsv'
@@ -285,7 +295,7 @@ def get_reset_val(value, width, lsb, xlen):
     return int(kBitSubStr, 2)
 
 
-def derive_warl_func(csr_name, width, warl_str):
+def derive_warl_func(field_name, width, warl, csr_dep):
     '''
         Function to convert the warl strings in the riscv-config YAML to
         corresponding bluespec code.
@@ -295,71 +305,85 @@ def derive_warl_func(csr_name, width, warl_str):
         separately to extract parameters and create bluespec code.
 
 
-        :param csr_name: string indicating the name of the csr
+        :param field_name: string indicating the name of the csr
         :param width: indicates the max size of the csr
         :param warl_str: containts the warl string from the riscv-config YAML string
 
-        :type csr_name: str
+        :type field_name: str
         :type width: int
         :type warl_str: str
     '''
 
-    warl_split = warl_str.split('->')
-    regex_range = re.compile(csr_name + '\[(.*?)\]\s*(.*?)\s*\[(.*?)\]')
-    regex_bitmask = re.compile(csr_name + '\[(.*?)\]\s*bitmask\s*\[(.*?)\]')
-    regex_cond = re.compile('(.*?)\[(.*?)\]\s*(.*?)\s*\[(.*?)\]')
     bsv_str = ''
-    condition_str = ''
-    if len(warl_split) > 1:
-        csr, indices, x, val = regex_cond.findall(warl_split[0])[0]
-        if 'not' in x:
-            eq = '=='
+    regex_range = re.compile(field_name + '\[(.*?)\]\s*(in|not in)\s*\[(.*?)\]')
+    regex_bitmask = re.compile(field_name + '\[(.*?)\]\s*bitmask\s*\[(.*?)\]')
+    regex_cond = re.compile('\s*(.*?)\[(.*?)\]\s*(.*?)\s*\[(.*?)\]')
+    for warl_str in warl:
+        warl_split = warl_str.split('->')
+        legal = warl_split[-1]
+        if csr_dep:
+            condition_str = 'if ('
         else:
-            eq = '!='
-        condition_str += 'if ( rg_{0}[{1}] {2} {3} )\n'.format(csr, indices, eq, val)
-        legal = warl_split[1]
-    else:
-        legal = warl_split[0]
-        condition = ''
-
-
-    range_vals = regex_range.findall(legal)
-    bitmask_vals = regex_bitmask.findall(legal)
-    full_warl = False
-    if range_vals != []:
-        bsv_str += 'if ( ('
-        for x in range_vals:
-            bsv_str += '('
-            indices = x[0]
-            if 'not' in x[1]:
-                not_eq = True
-            else:
-                not_eq = False
-            values = x[2].split(',')
-            for v in values:
-                if ':' in v:
-                    base_bound = v.split(':')
-                    if int(base_bound[0],16) == 0 and int(base_bound[1],16) == (2**int(width) -1):
-                        full_warl = True
-                    bsv_str += '( x[' + indices + '] >= ' +\
-                    base_bound[0] + ' && x[' + indices + '] <= ' +base_bound[1] + ') ||'
-                elif not_eq:
-                    bsv_str += '( x[' + indices + '] != ' + v + ') ||'
+            condition_str = ''
+        if csr_dep:
+            search = regex_cond.findall(warl_split[0])
+            for part in search:
+                csr, indices, op, val = part
+                if field_name == csr:
+                    csr_write = 'r'
                 else:
-                    bsv_str += '( x[' + indices + '] == ' + v + ') ||'
-            bsv_str = bsv_str[:-2] + ') && '
-        bsv_str = bsv_str[:-3] + ')'
-        bsv_str = bsv_str.replace('0x', "'h") + ')\n'
-        bsv_str += '                        r._write(x);'
-    elif bitmask_vals != []:
-        indices = bitmask_vals[0][0]
-        mask_fixed = bitmask_vals[0][1].split(',')
-        mask = mask_fixed[0][2:]
-        fixedval = mask_fixed[1][3:]
-        bsv_str = "r._write( (x & 'h" + mask + ") | (~'h" + mask + " & " + fixedval + ") );"
-    if full_warl:
-        bsv_str = 'r._write(x);'
-    bsv_str = condition_str + bsv_str
+                    for dep_csr in csr_dep:
+                        if csr in dep_csr and '::' in dep_csr:
+                            csr_write = 'rg_' + dep_csr.split('::')[0] + '_'+dep_csr.split('::')[1]
+                        elif csr == dep_csr:
+                            csr_write = 'rg_' + csr
+                op = '!=' if 'not in' in op else '=='
+                condition_str += ' {0}[{1}] {2} {3}'.format(csr_write, indices, op, val)
+                condition_str += ' && '
+            condition_str = condition_str[:-4] + ' )\n                  '
+        
+
+
+        range_vals = regex_range.findall(legal)
+        bitmask_vals = regex_bitmask.findall(legal)
+        full_warl = False
+        curr_bsv_str = ''
+        if range_vals != []:
+            curr_bsv_str += 'if ( ('
+            for x in range_vals:
+                curr_bsv_str += '('
+                indices = x[0]
+                if 'not' in x[1]:
+                    not_eq = True
+                else:
+                    not_eq = False
+                values = x[2].split(',')
+                for v in values:
+                    if ':' in v:
+                        base_bound = v.split(':')
+                        if int(base_bound[0],16) == 0 and int(base_bound[1],16) == (2**int(width) -1):
+                            full_warl = True
+                        curr_bsv_str += '( x[' + indices + '] >= ' +\
+                        base_bound[0] + ' && x[' + indices + '] <= ' +base_bound[1] + ') ||'
+                    elif not_eq:
+                        curr_bsv_str += '( x[' + indices + '] != ' + v + ') &&'
+                    else:
+                        curr_bsv_str += '( x[' + indices + '] == ' + v + ') ||'
+                curr_bsv_str = curr_bsv_str[:-2] + ') && '
+            curr_bsv_str = curr_bsv_str[:-3] + ')'
+            curr_bsv_str = curr_bsv_str.replace('0x', "'h") + ')\n'
+            curr_bsv_str += '                        r._write(x);'
+        elif bitmask_vals != []:
+            indices = bitmask_vals[0][0]
+            mask_fixed = bitmask_vals[0][1].split(',')
+            mask = mask_fixed[0][2:]
+            fixedval = mask_fixed[1][3:]
+            curr_bsv_str = "r._write( (x & 'h" + mask + ") | (~'h" + mask + " & " + fixedval + ") );"
+        if full_warl:
+            curr_bsv_str = 'r._write(x);'
+        bsv_str += condition_str + curr_bsv_str + '\n'
+        bsv_str += '                else  '
+    bsv_str = bsv_str[:-22]
     return bsv_str
 
 def station_interface(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
@@ -398,6 +422,13 @@ def station_interface(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
 
     if logging:
         logger.info('Initiating interface generation for : ' + pkg_name)
+    # define additional methods for certain csrs
+    req_regs = []
+    for csrname in isa_yaml:
+        if csrname.upper() in additional_ifc:
+            if additional_ifc[csrname.upper()] not in req_regs:
+                bsvstr += textwrap.indent(additional_ifc[csrname.upper()],'    ')
+                req_regs.append(additional_ifc[csrname.upper()])
 
     # define a value method for each csr.
     for csrname in isa_yaml:
@@ -502,12 +533,12 @@ def station_module(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
 
     if logging:
         logger.info('Creating module for : ' + pkg_name)
-
+    attr_str = ''
     for csrname in isa_yaml:
         if 'mod_attributes' in attr_yaml[csrname]:
             for modattr in attr_yaml[csrname]['mod_attributes']:
-                bsvstr += '  ' + modattr + '\n'
-    bsvstr += (module_temp.format(pkg_name))
+                attr_str += '  ' + modattr + '\n'
+    bsvstr += (module_temp.format(pkg_name, attr_str))
     if logging:
         logger.info('Module created for : ' + pkg_name)
 
@@ -542,7 +573,7 @@ def station_modbody (isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
         logger.info('Creating custom rules/module elements for : ' + pkg_name)
     for csrname in isa_yaml:
         if 'module_body' in attr_yaml[csrname]:
-            bsvstr += textwrap.indent(attr_yaml[csrname]['module_body'],'    ')
+            bsvstr += textwrap.indent(attr_yaml[csrname]['module_body'],'    ') + '\n'
     if logging:
         logger.info('Custom rules/module elements created for : ' + pkg_name)
     return bsvstr
@@ -576,6 +607,13 @@ def station_methods(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
     xlen = int(rvxlen[2:])
     if logging:
         logger.info('Creating method definitions for : ' + pkg_name)
+    # define additional methods for certain csrs
+    req_regs = []
+    for csrname in isa_yaml:
+        if csrname.upper() in additional_ifc_dec:
+            if additional_ifc_dec[csrname.upper()] not in req_regs:
+                bsvstr += textwrap.indent(additional_ifc_dec[csrname.upper()],'    ')
+                req_regs.append(additional_ifc_dec[csrname.upper()])
     
     # default value methods
     for csrname in isa_yaml:
@@ -768,6 +806,14 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
     bsvstr += (other_vars_temp)
     shadows = []
     xlen = int(rvxlen[2:])
+    # creating additional required registers
+    req_regs = []
+    for csrname in isa_yaml:
+        if csrname.upper() in additional_regs:
+            if additional_regs[csrname.upper()] not in req_regs:
+                bsvstr += textwrap.indent(additional_regs[csrname.upper()],'    ')
+                req_regs.append(additional_regs[csrname.upper()])
+
     # create wires of dependent csrs
     incomming_reads = []
     for csrname in isa_yaml:
@@ -819,7 +865,8 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
                         csrname, reset_val)
             elif ('warl' in content[rvxlen]['type']):
                 bsv_str = derive_warl_func( csrname, str(size),
-                        content[rvxlen]['type']['warl']['legal'][0])
+                        content[rvxlen]['type']['warl']['legal'], 
+                        content[rvxlen]['type']['warl']['dependency_fields'])
                 bsvstr += temp_warlReg.format(description, str(size), 
                             csrname, reset_val, bsv_str)
             else:
@@ -862,7 +909,8 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
                                 str(width), csrname+'_'+subfield, subreset_val)
                     elif 'warl' in type_reg:
                         bsv_str = derive_warl_func( subfield,
-                            str(width),subcontent['type']['warl']['legal'][0])
+                            str(width),subcontent['type']['warl']['legal'],
+                            subcontent['type']['warl']['dependency_fields'])
                         bsvstr += temp_warlReg.format(subdescription,
                                     str(width), csrname+'_'+subfield, subreset_val, bsv_str)
                     else:
@@ -922,7 +970,7 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     # initial license and imports
     bsvstr += license_header.format(version, str(datetime.datetime.now()))
     bsvstr += package_temp.format(pkgname)
-    
+
     misa_grp = find_group(grp_yaml, 'misa')
     mstatus_grp = find_group(grp_yaml, 'mstatus')
     mepc_grp = find_group(grp_yaml, 'mepc')
@@ -943,6 +991,9 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     mideleg_grp = find_group(grp_yaml, 'mideleg')
     sedeleg_grp = find_group(grp_yaml, 'sedeleg')
     sideleg_grp = find_group(grp_yaml, 'sideleg')
+    
+    pmp_entries = 0
+
 
     usertraps = False
     if None not in (ustatus_grp, uepc_grp, utval_grp, ucause_grp, utval_grp):
@@ -952,6 +1003,7 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
         supervisor = True
 
 
+
     # imports of each station
     if logging:
         logger.debug('-- Creating imports')
@@ -963,6 +1015,21 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     if logging:
         logger.debug('-- Creating interface')
     bsvstr += interface_temp.format(pkgname)
+    # define additional required interfaces
+    req_regs = []
+    for csrname in isa_yaml:
+        if csrname.upper() in additional_ifc:
+            if additional_ifc[csrname.upper()] not in req_regs:
+                bsvstr += textwrap.indent(additional_ifc[csrname.upper()],'    ')
+                req_regs.append(additional_ifc[csrname.upper()])
+        if 'pmpaddr' in csrname.lower():
+            if isa_yaml[csrname]['rv'+str(xlen)]['accessible']:
+                pmp_entries += 1
+
+    # pmp methods if required
+    if pmp_entries > 0:
+        bsvstr += pmp_ifc.format(pmp_entries, physical_addr_sz)
+
     # define a value method for each csr.
     for csrname in isa_yaml:
         if logging:
@@ -991,7 +1058,7 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     #--- create module
     if logging:
         logger.debug('-- Creating module')
-    bsvstr += module_temp.format(pkgname) + '\n'
+    bsvstr += module_temp.format(pkgname,'') + '\n'
 
     if logging:
         logger.debug('-- Instatiating groups')
@@ -1016,6 +1083,15 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     anyhit = anyhit[:-4]
     anydata = anydata[:-3]
     bsvstr += check_hit_temp.format(anyhit, anydata, xlen)
+    
+    # additional wires/regs
+    req_regs = []
+    for csrname in isa_yaml:
+        if csrname.upper() in additional_regs:
+            if additional_regs[csrname.upper()] not in req_regs:
+                bsvstr +=\
+                textwrap.indent(additional_regs[csrname.upper()],'    ') + '\n'
+                req_regs.append(additional_regs[csrname.upper()])
 
     # req=response connection across stations
     if logging:
@@ -1027,12 +1103,22 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     for grp in grp_yaml:
         bsvstr += temp_prv_connection.format(grp)
     
+    req_ifc = []
+    for csrname in isa_yaml:
+        if csrname.upper() in additional_topifc_dec:
+            grp_name = find_group(grp_yaml, csrname.upper())
+            if (grp_name, additional_topifc_dec[csrname.upper()]) not in req_ifc:
+                bsvstr +='\n'+\
+                textwrap.indent(additional_topifc_dec[csrname.upper()].format(grp_name),'    ')
+                req_ifc.append((grp_name,additional_topifc_dec[csrname.upper()]))
+    
     # create action methods of dependent csrs
     if logging:
         logger.debug('-- defining action methods')
     for grp in grp_yaml:
         incomming_reads = []
-        for csrs in grp_yaml[grp]:
+        if grp_yaml[grp] != None :
+         for csrs in grp_yaml[grp]:
             csrname = csrs.lower()
             if 'depends_on_csr' in attr_yaml[csrname]:
                 if len(attr_yaml[csrname]['depends_on_csr']) != 0:
@@ -1101,6 +1187,43 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     bsvstr += method_top_trap_machine_temp.format(mtval_grp, mepc_grp, mcause_grp,
             mstatus_grp, mtvec_grp)
     
+    # define additional required interfaces
+    req_ifc = []
+    for csrname in isa_yaml:
+        if csrname.upper() in additional_ifc_dec:
+            if additional_ifc_dec[csrname.upper()] not in req_ifc:
+                bsvstr +=\
+                textwrap.indent(additional_ifc_dec[csrname.upper()],'    ') + '\n'
+                req_ifc.append(additional_ifc_dec[csrname.upper()])
+
+    # pmp method definitions
+    pmpaddr_str = ''
+    pmpcfg_str = ''
+    cfg_count = 0
+    count = 0
+    if pmp_entries > 0:
+        for csrname in isa_yaml:
+            if 'pmpaddr' in csrname.lower():
+                num = int(re.findall('\d+',csrname.lower())[0])
+                grp = find_group(grp_yaml, csrname.upper())
+                pmpaddr_str += "        lv_pmpaddr[{0}] = {{truncate({1}.mv_csr_{2}),2'b0}};\n".format(\
+                        count, grp, csrname.lower())
+                count += 1
+
+                # cgfs are notoriously ordered and hence the following logic
+                cfg_num = (num // (xlen//8))
+                if xlen == 64 and (cfg_num%2) == 1:
+                    cfg_num = cfg_num + 1
+                cfg_offset = int(num%(xlen//8))
+                cfg_grp = find_group(grp_yaml, 'PMPCFG'+str(cfg_num))
+                pmpcfg_str += \
+            '        lv_pmpcfg[{0}] = {1}.mv_csr_pmpcfg{2}[{3}:{4}];\n'.format(\
+                    cfg_count, cfg_grp, cfg_num, cfg_offset*8+7,
+                    cfg_offset*8)
+                cfg_count += 1
+        bsvstr += pmpaddr_def.format(pmp_entries, physical_addr_sz, pmpaddr_str)
+        bsvstr += pmpcfg_def.format(pmp_entries, pmpcfg_str)
+
     # complete the value method for each csr.
     if logging:
         logger.debug('-- Defining value methods')
